#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <omp.h>
#include <mpi.h>
#include <signal.h>
#include <unistd.h>

// Ejemplo de ordenación por el método de la burbuja
//esquema para la entrada/salida, validación y toma de tiempos
//para el proyecto de Iniciación al paralelismo en módulos de FP
//Universidad de Murcia-Instituto de Enseñanza Secundaria Ingeniero de la Cierva, Murcia, 2014

void generar(double *x,int t)
//se generan t datos reales con valores entre -10 y 10
{
  int i;

  for(i=0;i<t;i++)
  {  
     x[i] =((1.*rand())/RAND_MAX)*20-10;
  }
}

void escribir(double *x,int t)
{
  int i;

  for(i=0;i<t;i++)
  {  
     printf("%.4lf ",x[i]);
  }
  printf("\n");
}

void escribirresult(double *a,int N,int salida)
{
  int i;

  for(i=0;i<N;i++)
  {
        if((i%salida)==0)
        {
          printf("%.4lf ",a[i]);
        }
  }
  printf("\n");
}

/*
c
c     mseconds - returns elapsed milliseconds since Jan 1st, 1970.
c
*/
long long mseconds(){
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec*1000 + t.tv_usec/1000;
}

static void alarm_handler(int sig) {
  fprintf(stderr, "Time Limit Exceeded\n");
  abort();
}

extern void sec(double *a,int t,int nodo,int np);

int main(int argc,char *argv[]) {
  int nodo,np;
  int i,j,n;
  int cuantos,semilla,salida;
  long long ti,tf,tt=0;
  double *datos;

  FILE *stats_file = fopen("stats", "w");

  struct sigaction sact;
  sigemptyset(&sact.sa_mask);
  sact.sa_flags = 0;
  sact.sa_handler = alarm_handler;
  sigaction(SIGALRM, &sact, NULL);
  alarm(60);  /* time limit */
 
  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&np);
  MPI_Comm_rank(MPI_COMM_WORLD,&nodo);

// The number of test cases is read
// and sent to all the processes
  if(nodo==0)
  {
    scanf("%d",&cuantos);
    MPI_Bcast(&cuantos,1,MPI_INT,0,MPI_COMM_WORLD);
  }
  else
  {
    MPI_Bcast(&cuantos,1,MPI_INT,0,MPI_COMM_WORLD);
  }
  for(i=0;i<cuantos;i++)
  {
    if(nodo==0)
    {
       scanf("%d",&n);                   // Tamaño del array
       scanf("%d",&semilla);             // semilla para la generación aleatoria
       scanf("%d",&salida);                 // para escribir valores seleccionados de la solución

// espacio para los datos
       datos = (double *)malloc(sizeof(double)*n);

      srand(semilla);
      generar(datos,n);
    }
    MPI_Barrier(MPI_COMM_WORLD);
    ti=mseconds(); 
    sec(datos,n,nodo,np);
    MPI_Barrier(MPI_COMM_WORLD);
    tf=mseconds(); 
    if(nodo==0)
    {
// The time of the first input is not considered
      if(i!=0) tt+=tf-ti;
#ifdef DEBUG
// While debugging the time of each execution is written
      fprintf(stats_file, "%Ld\n", tf-ti);
#endif
// The results of each problem are written,
// even for the first problem
      escribirresult(datos,n,salida);
      free(datos);
    } 
  }
  
  if(nodo==0)
  { 
    fprintf(stats_file, "%Ld\n", tt);
    fclose(stats_file);
  }
  MPI_Finalize();
  return 0;
}
