/*
  CPP_CONTEST=PP-FP2014
  CPP_PROBLEM=C
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 1
*/

#include <stdlib.h>
#include <stdio.h>
#include <omp.h>
#include <mpi.h>
 
#define DISTRIBUCION 10
#define COMBINACION 20

void mezclar(double *a, int t) {
  int i, j, k;
  double *b = (double *) malloc(sizeof(double) * t * 2);

  i = 0;
  j = t;
  k = 0;
  while (i < t && j < 2 * t) {
    if (a[i] < a[j]) {
      b[k] = a[i];
      i++;
    } else {
      b[k] = a[j];
      j++;
    }
    k++;
  }
  for (; i < t; i++) {
    b[k] = a[i];
    k++;
  }  
  for (; j < 2 * t; j++) {
    b[k] = a[j];
    k++;
  }      
  for (i = 0; i < 2 * t; i++)
    a[i] = b[i];
  free(b);
}

void mergesortsec(double *a, int t) {
  if (t != 1) {
    mergesortsec(a, t / 2);
    mergesortsec(&a[t / 2], t / 2);
    mezclar(a, t / 2);
  }
}

// Código 6.16
      
void distribuir(double *a, int n, int nodo, int np) {
    int des = 1;
    int tam = n / 2;
// se hace una distribución en árbol
    while (tam >= n / np) {
        if (nodo < des)
            MPI_Send(&a[tam], tam, MPI_DOUBLE, nodo + des, DISTRIBUCION, MPI_COMM_WORLD);
        if (nodo >= des && nodo < 2 * des)
            MPI_Recv(a, tam, MPI_DOUBLE, nodo - des, DISTRIBUCION, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        des *= 2;
        tam /= 2;
    }
}

void combinar(double *a, int n, int nodo, int np) {
    int des = np / 2;
    int tam = n / np;
// la combinación se hace en árbol
    while (tam < n) {
        if (nodo >= des && nodo < 2 * des)
        MPI_Send(a, tam, MPI_DOUBLE, nodo - des, COMBINACION, MPI_COMM_WORLD);
        if (nodo < des) {
            MPI_Recv(&a[tam], tam, MPI_DOUBLE, nodo + des, COMBINACION, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
// realizando mezclas los procesos que reciben datos
            mezclar(a, tam);
        }
        des /= 2;
        tam *= 2;
    }
}

void sec(double *a, int n, int nodo, int np) {
    MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
    if(nodo!=0)
    {
        a=(double *) malloc(sizeof(double)*n);
    }
    distribuir(a, n, nodo, np);
    mergesortsec(a, n / np);
    combinar(a, n, nodo, np);
    if(nodo!=0)
        free(a);
}

