//Proyecto de Iniciación a la programación paralela en los módulos de informática
//Universidad de Murcia - Instituto de Enseñanza Secundaria Ingeniero de la Cerva, Murcia
//2014

//Ejemplo de paralelismo con OpenMP
//Ordenación utilizando varios hilos

//En linux, 
//compilar con: g++ ejemploOMPordenacion.cpp -O3 -fopenmp 
//y la -D DEBUG si se quiere compilar la zona de escritura de datos
//ejecutar con: ./a.out
//e introducir el número de datos que se quiere ordenar

//Hay que incluir la librería de OpenMP
#include <omp.h>

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <time.h>
using namespace std;

//esta constante se usará para el número de hilos que se ponen en marcha
#define NUM_THREADS 2

void generar(float *x,int t)
//se generan t datos reales con valores entre -10 y 10
{
  int i;

  for(i=0;i<t;i++)
  {  
     x[i] =((1.*rand())/RAND_MAX)*20-10;
  }
}

void escribir(float *x,int t)
{
  int i;

  for(i=0;i<t;i++)
  {  
     cout <<x[i]<< "  ";
  }
  cout <<endl;
}

void ordenar(float *x,int t)
{
    int i,j;
    float temp;

    for(i=0;i<t;i++)
    {
        for(j=i+1;j<t;j++)
        {
             if(x[i]>x[j])
             {
                 temp=x[i];
                 x[i]=x[j];
                 x[j]=temp;
             }
         }
     }
}

//Se ha modificado la mezcla porque al usar más de dos hilos 
//hay que hacer varias mezclas en zonas distintas
void mezclar(float *x,int izq,int med,int der)
{
    int i,j,k;
    float *temp=(float *) malloc(sizeof(float)*(der-izq));

    i=izq;
    j=med;
    k=0;
    while(i<med && j<der)
    {
       if(x[i]<x[j])
       {
            temp[k]=x[i];
            i++;
       }
       else
       {
            temp[k]=x[j];
            j++;
       }
       k++;
    }
    for(;i<med;i++)
    {
         temp[k]=x[i];
         k++;
    }
    for(;j<der;j++)
    {
        temp[k]=x[j];
        k++;
    }
    for(i=0;i<der-izq;i++)
    {
        x[i+izq]=temp[i];
    }
    free(temp);
}
    
void copiar(float *x,float *y,int t)
{
    int i;
    
    for(i=0;i<t;i++)
    {
        y[i]=x[i];
    }
}

float comparar(float *x,float *y,int t)
{
    int i;
    float total=0;

    for(i=0;i<t;i++)
    {
        total=(y[i]>x[i]?y[i]-x[i]:x[i]-y[i]);
    }
    return total;
}

long long mseconds(){
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec*1000 + t.tv_usec/1000;
}

int main()
{
    int n;
    float *datos,*copiadatos;
    long long ti,tf;

    cin >>n; //se lee el número de datos que se van a ordenar

    datos = (float *)malloc(sizeof(float)*n);
    copiadatos=(float *)malloc(sizeof(float)*n);

    generar(datos,n);
    copiar(datos,copiadatos,n);

//se ordena la copia para comparar el resultado de la ordenación secuencial con la paralela
//y se toman tiempos para comparar los tiempos de las dos versiones
    ti=mseconds(); 
    ordenar(copiadatos,n);
    tf=mseconds();
    
    cout <<"Tiempo secuencial: " <<tf-ti<<" microsegundos"<<endl;

//Si se compila con -D DEBUG se escriben los datos
#ifdef DEBUG
    escribir(datos,n);
#endif

//toma de tiempos de la ordenación paralela
   ti=mseconds(); 

//Se indica el número de hilos que trabajarán en una región paralela
    omp_set_num_threads(NUM_THREADS);

//el hilo maestro pone en marcha los esclavos
#pragma omp parallel 
{
//los hilos obtienen su identificador, que va de 0 a NUM_THREADS-1
    int nodo=omp_get_thread_num();
    if(nodo!=NUM_THREADS-1)
    {
//todos los hilos trabajan con la misma cantidad de datos
        ordenar(&datos[nodo*(n/NUM_THREADS)],n/NUM_THREADS);
    }
    else
    {
//salvo el último, que ordena todos los datos restantes
        ordenar(&datos[nodo*(n/NUM_THREADS)],n-nodo*(n/NUM_THREADS));
    }
}    
//La sincronización entre hilos se hace al salir de la región paralela
//a partir de aquí trabaja solo el hilo maestro

#ifdef DEBUG
    escribir(datos,n);
#endif

//El hilo maestro mezcla los trozos ordenados.
//Se podría hacer de forma más eficiente, mezclando cada vez trozos de tamaño similar,
//e incluso usando hilos distintos para mezclas distintas
   for(int i=NUM_THREADS-2;i>=0;i--)
   {
       int izq=i*(n/NUM_THREADS);
       int med=izq+n/NUM_THREADS;
       mezclar(datos,izq,med,n);
#ifdef DEBUG
    escribir(datos,n);
#endif
   }
#ifdef DEBUG
    escribir(datos,n);
#endif
    tf=mseconds();

    cout <<"Tiempo paralelo: " <<tf-ti<<" microsegundos"<<endl;
    cout <<"Diferencia: "<<comparar(datos,copiadatos,n)<<endl;

    pthread_exit(NULL);
}

