//Proyecto de Iniciación a la programación paralela en los módulos de informática
//Universidad de Murcia - Instituto de Enseñanza Secundaria Ingeniero de la Cerva, Murcia
//2014

//Ejemplo de uso de pthreads
//Ordenación utilizando varios hilos

//En linux, 
//compilar con: g++ ejemplothreads.cpp 
//incluir la opción -O3 para optimizar el código: g++ -O3 ejemplofork.cpp -fpermissive -lpthread
// -lfpermissive para que admita una conversión en el paso de parámetros
//  -lpthread para usar la librería pthreads
//y la -D DEBUG si se quiere compilar la zona de escritura de datos
//ejecutar con: ./a.out
//e introducie el número de datos que se quiere ordenar

//Hay que incluir la librería de pthreads
#include <pthread.h>

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <time.h>
using namespace std;

//esta constante se usará para el número de hilos que se ponen en marcha
#define NUM_THREADS 2

//estructura que se pasa a cada hilo
//tamaño de datos con que tiene que trabajar y zona de comienzo de los datos
   struct thread_data{
     int  t;
     float *datos;
  };

void generar(float *x,int t)
//se generan t datos reales con valores entre -10 y 10
{
  int i;

  for(i=0;i<t;i++)
  {  
     x[i] =((1.*rand())/RAND_MAX)*20-10;
  }
}

void escribir(float *x,int t)
{
  int i;

  for(i=0;i<t;i++)
  {  
     cout <<x[i]<< "  ";
  }
  cout <<endl;
}

//esta ordenación es la normal y se usará para la ordenación secuencial
void ordenarsec(float *x,int t)
{
    int i,j;
    float temp;

    for(i=0;i<t;i++)
    {
        for(j=i+1;j<t;j++)
        {
             if(x[i]>x[j])
             {
                 temp=x[i];
                 x[i]=x[j];
                 x[j]=temp;
             }
         }
     }
}

//esta es la función de ordenar que se dirá a cada hilo que haga
//tiene que ser una referencia a una función
//y se le pasa la dirección de memoria de la estructura donde se ha puesto el tamaño y la dirección de los datos con los que trabajar
void *ordenar(struct thread_data *entrada)
{
    int t=entrada->t;
    float *x=entrada->datos;
    int i,j;
    float temp;

    for(i=0;i<t;i++)
    {
        for(j=i+1;j<t;j++)
        {
             if(x[i]>x[j])
             {
                 temp=x[i];
                 x[i]=x[j];
                 x[j]=temp;
             }
         }
     }
}

//Se ha modificado la mezcla porque al usar más de dos hilos 
//hay que hacer varias mezclas en zonas distintas
void mezclar(float *x,int izq,int med,int der)
{
    int i,j,k;
    float *temp=(float *) malloc(sizeof(float)*(der-izq));

    i=izq;
    j=med;
    k=0;
    while(i<med && j<der)
    {
       if(x[i]<x[j])
       {
            temp[k]=x[i];
            i++;
       }
       else
       {
            temp[k]=x[j];
            j++;
       }
       k++;
    }
    for(;i<med;i++)
    {
         temp[k]=x[i];
         k++;
    }
    for(;j<der;j++)
    {
        temp[k]=x[j];
        k++;
    }
    for(i=0;i<der-izq;i++)
    {
        x[i+izq]=temp[i];
    }
    free(temp);
}
    
void copiar(float *x,float *y,int t)
{
    int i;
    
    for(i=0;i<t;i++)
    {
        y[i]=x[i];
    }
}

float comparar(float *x,float *y,int t)
{
    int i;
    float total=0;

    for(i=0;i<t;i++)
    {
        total=(y[i]>x[i]?y[i]-x[i]:x[i]-y[i]);
    }
    return total;
}

long long mseconds(){
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec*1000 + t.tv_usec/1000;
}

int main()
{
    int n;
    float *datos,*copiadatos;
    long long ti,tf;
    void *status; //se usa para comprobar el estado de los hilos
    pthread_t threads[NUM_THREADS]; //se declara un array de hilos para ponerlos en marcha

    cin >>n; //se lee el número de datos que se van a ordenar

    datos = (float *)malloc(sizeof(float)*n);
    copiadatos=(float *)malloc(sizeof(float)*n);

    generar(datos,n);
    copiar(datos,copiadatos,n);

//se ordena la copia para comparar el resultado de la ordenación secuencial con la paralela
//y se toman tiempos para comparar los tiempos de las dos versiones
    ti=mseconds(); 
    ordenarsec(copiadatos,n);
    tf=mseconds();
    
    cout <<"Tiempo secuencial: " <<tf-ti<<" microsegundos"<<endl;

//Si se compila con -D DEBUG se escriben los datos
#ifdef DEBUG
    escribir(datos,n);
#endif

//toma de tiempos de la ordenación paralela
   ti=mseconds(); 

//hay que declarar tantas estructuras como hilos
   struct thread_data thread_data_array[NUM_THREADS];
//en este bucle se crearán los hilos
   for(int i=0; i<NUM_THREADS; i++){
//todos los hilos trabajan con la misma cantidad de datos
      thread_data_array[i].t=n/NUM_THREADS;
//salvo el último, que ordena todos los datos restantes
      if(i==(NUM_THREADS-1))
      {
           thread_data_array[i].t=n-i*(n/NUM_THREADS);
      }
      thread_data_array[i].datos=&datos[i*(n/NUM_THREADS)];
//se crea el hilo pasando su dirección, la función que ejecuta, y los parámetros de la función en una estructura
      pthread_create(&threads[i], NULL, ordenar,(void *) &thread_data_array[i]);
   }

//el hilo maestro espera a que los esclavos acaben
   for(int i=0; i<NUM_THREADS; i++) {
      pthread_join(threads[i], &status);
   }

#ifdef DEBUG
    escribir(datos,n);
#endif

//El hilo maestro mezcla los trozos ordenados por los esclavos
//Hay tantos trozos como hilos esclavos y se realiza una mezcla menos que trozos hay.
//Se podría hacer de forma más eficiente, mezclando cada vez trozos de tamaño similar,
//e incluso usando hilos distintos para mezclas distintas
   for(int i=NUM_THREADS-2;i>=0;i--)
   {
       int izq=i*(n/NUM_THREADS);
       int med=izq+n/NUM_THREADS;
       mezclar(datos,izq,med,n);
#ifdef DEBUG
    escribir(datos,n);
#endif
   }
#ifdef DEBUG
    escribir(datos,n);
#endif
    tf=mseconds();

    cout <<"Tiempo paralelo: " <<tf-ti<<" microsegundos"<<endl;
    cout <<"Diferencia: "<<comparar(datos,copiadatos,n)<<endl;

    pthread_exit(NULL);
}

